/*
 * Copyright (c) 2013, Victor Nazarov <asviraspossible@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Victor Nazarov nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.github.sviperll;

import java.util.ArrayList;
import java.util.List;

public class Predicates {
    public static <T> ConfigurableFunction<T, Boolean> asFunction(Predicate<T> p) {
        return new ConfigurableFunction<T, Boolean>(new PredicateFunction<T>(p));
    }

    public static <T> ConfigurablePredicate<T> not(Predicate<T> p) {
        return new ConfigurablePredicate<T>(new NotPredicate<T>(p));
    }

    public static <T> ConfigurablePredicate<T> and(Predicate<T> a, Predicate<T> b) {
        List<Predicate<T>> predicates = new ArrayList<Predicate<T>>(2);
        predicates.add(a);
        predicates.add(b);
        return new ConfigurablePredicate<T>(new AndPredicate<T>(predicates));
    }

    public static <T> ConfigurablePredicate<T> or(Predicate<T> a, Predicate<T> b) {
        List<Predicate<T>> predicates = new ArrayList<Predicate<T>>(2);
        predicates.add(a);
        predicates.add(b);
        return new ConfigurablePredicate<T>(new OrPredicate<T>(predicates));
    }

    public static <T> ConfigurablePredicate<T> and(List<Predicate<T>> predicates) {
        return new ConfigurablePredicate<T>(new AndPredicate<T>(predicates));
    }

    public static <T> ConfigurablePredicate<T> or(List<Predicate<T>> predicates) {
        return new ConfigurablePredicate<T>(new OrPredicate<T>(predicates));
    }

    public static <T> ConfigurablePredicate<T> configure(Predicate<T> predicate) {
        if (predicate instanceof ConfigurablePredicate)
            return (ConfigurablePredicate<T>)predicate;
        else
            return new ConfigurablePredicate<T>(predicate);
    }

    private Predicates() {}

    private static class NotPredicate<T> implements Predicate<T> {
        private final Predicate<T> p;
        public NotPredicate(Predicate<T> p) {
            this.p = p;
        }

        @Override
        public boolean evaluate(T t) {
            return !p.evaluate(t);
        }
    }

    private static class AndPredicate<T> implements Predicate<T> {
        private final List<Predicate<T>> predicates;
        public AndPredicate(List<Predicate<T>> predicates) {
            this.predicates = predicates;
        }

        @Override
        public boolean evaluate(T t) {
            for (Predicate<T> predicate: predicates)
                if (!predicate.evaluate(t))
                    return false;
            return true;
        }
    }

    private static class OrPredicate<T> implements Predicate<T> {
        private final List<Predicate<T>> predicates;
        public OrPredicate(List<Predicate<T>> predicates) {
            this.predicates = predicates;
        }

        @Override
        public boolean evaluate(T t) {
            for (Predicate<T> predicate: predicates)
                if (predicate.evaluate(t))
                    return true;
            return false;
        }
    }

    private static class PredicateFunction<T> implements Function<T, Boolean> {
        private final Predicate<T> predicate;
        public PredicateFunction(Predicate<T> predicate) {
            this.predicate = predicate;
        }

        @Override
        public Boolean apply(T t) {
            return predicate.evaluate(t);
        }
    }
}
