/*
 * Copyright (c) 2013, Victor Nazarov <asviraspossible@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Victor Nazarov nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.github.sviperll.collection;

import com.github.sviperll.BiFunction;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import com.github.sviperll.Function;
import com.github.sviperll.Predicate;

public interface FluentIterator<T> extends Iterator<T> {
    /**
     * @param iterator another iterator to append after this
     * @returns iterator wich contains elements from this iterator and from another iterator
     */
    FluentIterator<T> append(Iterator<? extends T> iterator);

    /**
     * @param function function to map each element
     * @returns iterator wich is the same as original, but contains not the original
     *                   elements of this iterator, but the result of applying function
     *                   to each element
     */
    <U> FluentIterator<U> map(Function<? super T, ? extends U> function);

    /**
     * @param predicate predicate to check elements whether each element should stay or will be
     *                  discarded
     * @returns iterator wich is the same as original, but contains only those elements
     *                   for wich predicate is true
     */
    FluentIterator<T> filter(Predicate<? super T> predicate);

    /**
     * @param count number of first elements to drop
     * @returns iterator wich is the same as original, but do not include count first elements
     *                   of original iterator
     */
    FluentIterator<T> drop(int count);

    /**
     * @param count number of first elements to take
     * @returns iterator wich is the same as original, but have only count first elements
     */
    FluentIterator<T> take(int count);

    /**
     * @param seed seed value
     * @param combiner operation to combine
     * @returns iterator wich is the same as original, but have only count first elements
     */
    <U> U fold(U seed, BiFunction<? super T, U, U> combiner);

    /**
     * @returns new array list filled with elements of this iterator
     */
    List<T> intoNewArrayList();

    /**
     * @returns new tree set filled with elements of this iterator
     */
    Set<T> intoNewTreeSet();

    /**
     * @returns new hash set filled with elements of this iterator
     */
    Set<T> intoNewHashSet();
}
