/*
 * Copyright (c) 2012, Victor Nazarov <asviraspossible@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Victor Nazarov nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.github.sviperll.collection;

import com.github.sviperll.BiFunction;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import com.github.sviperll.Function;
import com.github.sviperll.Predicate;
import com.github.sviperll.Functions;
import com.github.sviperll.collection.iterable.EmptyIterable;
import com.github.sviperll.collection.iterable.JoinedIterable;
import com.github.sviperll.collection.iterable.FilteredIterable;
import com.github.sviperll.collection.iterable.IterableFluentIterable;
import com.github.sviperll.collection.iterable.MappedIterable;
import com.github.sviperll.collection.iterable.PairIterable;
import com.github.sviperll.collection.iterable.SingletonIterable;
import com.github.sviperll.collection.iterable.TakenIterable;
import com.github.sviperll.collection.iterable.ThreeAppendedIterablesIterable;
import com.github.sviperll.collection.iterable.TripleIterable;
import com.github.sviperll.collection.iterable.TwoAppendedIterablesIterable;

public class Iterables {
    private static final Function ITERATOR_FUNCTION = new IteratorFunction();

    /**
     * Convert given iterable to the type of FluentIterable.
     * This convertion decorates original iterable with many new useful methods.
     * FluentIterable provides much richer API (@see FluentIterable)
     *
     * @param iterable iterable to convert
     * @returns decorated iterable of type FluentIterable.
     */
    public static <T> FluentIterable<T> decorate(Iterable<T> iterable) {
        if (iterable instanceof FluentIterable)
            return (FluentIterable<T>)iterable;
        else
            return new IterableFluentIterable<T>(iterable);
    }

    /**
     * @returns Mapper that gets iterator from given iterable
     */
    @SuppressWarnings("unchecked")
    public static <T> Function<? super Iterable<? extends T>, Iterator<? extends T>> iteratorFunction() {
        return ITERATOR_FUNCTION;
    }

    /**
     * @returns Iterable wich contains no elements
     */
    public static <T> FluentIterable<T> empty() {
        return decorate(new EmptyIterable<T>());
    }

    /**
     * @param element1 element to include into resulting iterable
     * @returns Iterable wich contains one element
     */
    public static <T> FluentIterable<T> iterableOf(T element1) {
        return decorate(new SingletonIterable<T>(element1));
    }

    /**
     * @param element1 first element to include into resulting iterable
     * @param element2 second element to include into resulting iterable
     * @returns Iterable wich contains two elements
     */
    public static <T> FluentIterable<T> iterableOf(T element1, T element2) {
        return decorate(new PairIterable<T>(element1, element2));
    }

    /**
     * @param element1 first element to include into resulting iterable
     * @param element2 second element to include into resulting iterable
     * @param element3 second element to include into resulting iterable
     * @returns Iterable wich contains two elements
     */
    public static <T> FluentIterable<T> iterableOf(T element1, T element2, T element3) {
        return decorate(new TripleIterable<T>(element1, element2, element3));
    }

    /**
     * @param iterable original iterable to base resulting iterable upon
     * @param function to apply to each allement of original iterable
     * @returns iterable wich the same as original iterable, but contains not original
     *          elements, but a result of applying mapper to each element of original iterable
     */
    public static <T, R> FluentIterable<R> map(Iterable<T> iterable, Function<? super T, ? extends R> function) {
        return decorate(new MappedIterable<T, R>(iterable, function));
    }

    /**
     * @param iterable original iterable to base resulting iterable upon
     * @param predicate to check wheather to leave element or not
     * @returns iterable wich the same as original iterable, but contains only those elements
     *          of original iterable for which predicate evaluates to true
     */
    public static <E, F> FluentIterable<E> filter(Iterable<E> iterable, Predicate<? super E> predicate) {
        return decorate(new FilteredIterable<E>(iterable, predicate));
    }

    /**
     * @param iterable iterable to base result upon
     * @param combiner operation to combine each element of iterable with seed
     * @param seed value to base result upon
     * @returns the result of applying combiner to the seed and to each element of iterable.
     */
    public static <E, F> F fold(Iterable<? extends E> eiterable, F seed, BiFunction<? super E, F, F> combiner) {
        return Iterators.decorate(eiterable.iterator()).fold(seed, combiner);
    }

    /**
     * @param iterable1 first iterable
     * @param iterable2 second iterable to append to the first
     * @returns iterable wich contains elements from first iterable and from second in this order.
     */
    public static <T> FluentIterable<T> append(Iterable<? extends T> iterable1, Iterable<? extends T> iterable2) {
        return decorate(new TwoAppendedIterablesIterable<T>(iterable1, iterable2));
    }

    /**
     * @param iterable1 first iterable
     * @param iterable2 second iterable to append to the first
     * @param iterable2 third iterable to append to the first and second
     * @returns iterable wich contains elements from first, second and third iterable in this order.
     */
    public static <T> FluentIterable<T> append(Iterable<? extends T> iterable1, Iterable<? extends T> iterable2, Iterable<? extends T> iterable3) {
        return decorate(new ThreeAppendedIterablesIterable<T>(iterable1, iterable2, iterable3));
    }

    /**
     * @param iterable wich contains iterables
     * @returns iterable wich contains elements from each iterables contained in original iterable.
     */
    public static <T> FluentIterable<T> join(Iterable<? extends Iterable<? extends T>> itrables) {
        return decorate(new JoinedIterable<T>(itrables));
    }

    /**
     * @param iterable to take elements from
     * @param count number of elements to take
     * @returns iterable wich contains only first count elements from iterable.
     */
    public static <T> FluentIterable<T> take(Iterable<T> itrable, int count) {
        return decorate(new TakenIterable<T>(itrable, count));
    }

    /**
     * @param iterable to take elements from
     * @param count number of elements to take
     * @returns iterable wich contains all except first count elements from iterable.
     */
    public static <T> FluentIterable<T> drop(Iterable<T> itrable, int count) {
        return decorate(new TakenIterable<T>(itrable, count));
    }

    /**
     * @param iterable iterable to take element from
     * @returns new array list wich contain all elements of this iterable.
     */
    public static <T> List<T> intoNewArrayList(Iterable<? extends T> iterable) {
        List<T> list = Collections.newArrayList();
        for (T t: iterable)
            list.add(t);
        return list;
    }

    /**
     * @param iterable iterable to take element from
     * @returns new tree set wich contain all elements of this iterable.
     */
    public static <T> Set<T> intoNewTreeSet(Iterable<? extends T> iterable) {
        Set<T> set = Collections.newTreeSet();
        for (T t: iterable)
            set.add(t);
        return set;
    }

    /**
     * @param iterable iterable to take element from
     * @returns new hash set wich contain all elements of this iterable.
     */
    public static <T> Set<T> intoNewHashSet(Iterable<? extends T> iterable) {
        Set<T> set = Collections.newHashSet();
        for (T t: iterable)
            set.add(t);
        return set;
    }

    private Iterables() {}

    private static class IteratorFunction<T> implements Function<Iterable<? extends T>, Iterator<? extends T>> {
        @Override
        public Iterator<? extends T> apply(Iterable<? extends T> t) {
            return t.iterator();
        }
    }
}
