/*
 * Copyright (c) 2012, Victor Nazarov <asviraspossible@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Victor Nazarov nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.github.sviperll.collection;

import com.github.sviperll.BiFunction;
import java.util.ArrayDeque;
import java.util.Iterator;
import java.util.List;
import java.util.Queue;
import java.util.Set;
import com.github.sviperll.Function;
import com.github.sviperll.Predicate;
import com.github.sviperll.collection.iterator.IteratorFluentIterator;
import com.github.sviperll.collection.iterator.SingletonIterator;
import com.github.sviperll.collection.iterator.EmptyIterator;
import com.github.sviperll.collection.iterator.FilteredIterator;
import com.github.sviperll.collection.iterator.JoinedIterator;
import com.github.sviperll.collection.iterator.MappedIterator;
import com.github.sviperll.collection.iterator.QueueIterator;
import com.github.sviperll.collection.iterator.TakenIterator;

public class Iterators {
    @SuppressWarnings("unchecked")
    private static final FluentIterator EMPTY = new IteratorFluentIterator(new EmptyIterator());

    /**
     * Convert given iterator to the type of FluentIterator.
     * This convertion decorates original iterator with many new useful methods.
     * FluentIterator provides much richer API (@see FluentIterable)
     *
     * @param iterator iterator to convert
     * @returns decorated iterator of type FluentIterator.
     */
    public static <T> FluentIterator<T> decorate(Iterator<T> iterator) {
        if (iterator instanceof FluentIterator)
            return (FluentIterator<T>)iterator;
        else
            return new IteratorFluentIterator<T>(iterator);
    }

    /**
     * @param iterator original iterator to base resulting iterator upon
     * @param function to apply to each allement of original iterator
     * @returns iterator wich the same as original iterator, but contains not original
     *          elements, but a result of applying mapper to each element of original iterator
     */
    public static <T, R> FluentIterator<R> map(Iterator<T> source, Function<? super T, ? extends R> function) {
        return decorate(new MappedIterator<T, R>(decorate(source), function));
    }

    /**
     * @param iterator original iterator to base resulting iterator upon
     * @param predicate to check wheather to leave element or not
     * @returns iterator wich the same as original iterator, but contains only those elements
     *          of original iterator for which predicate evaluates to true
     */
    public static <T> FluentIterator<T> filter(Iterator<T> source, Predicate<? super T> predicate) {
        return decorate(new FilteredIterator<T>(decorate(source), predicate));
    }

    /**
     * @returns Iterator wich contains no elements
     */
    @SuppressWarnings("unchecked")
    public static <T> FluentIterator<T> empty() {
        return EMPTY;
    }

    /**
     * @param element1 element to include into resulting iterator
     * @returns Iterator wich contains one element
     */
    public static <T> FluentIterator<T> iteratorOf(T value) {
        return decorate(new SingletonIterator<T>(value));
    }

    /**
     * @param element1 first element to include into resulting iterator
     * @param element2 second element to include into resulting iterator
     * @returns iterator wich contains two elements
     */
    public static <T> FluentIterator<T> iteratorOf(T value1, T value2) {
        Queue<T> queue = new ArrayDeque<T>(2);
        queue.add(value1);
        queue.add(value2);
        return decorate(new QueueIterator<T>(queue));
    }

    /**
     * @param element1 first element to include into resulting iterator
     * @param element2 second element to include into resulting iterator
     * @param element3 second element to include into resulting iterator
     * @returns iterator wich contains two elements
     */
    public static <T> FluentIterator<T> iteratorOf(T value1, T value2, T value3) {
        Queue<T> queue = new ArrayDeque<T>(2);
        queue.add(value1);
        queue.add(value2);
        queue.add(value3);
        return decorate(new QueueIterator<T>(queue));
    }

    /**
     * @param iterator1 first iterator
     * @param iterator2 second iterator to append to the first
     * @returns iterator wich contains elements from first iterator and from second in this order.
     */
    public static <T> FluentIterator<T> append(Iterator<? extends T> iterator1, Iterator<? extends T> iterator2) {
        Queue<FluentIterator<? extends T>> queue = new ArrayDeque<FluentIterator<? extends T>>(2);
        queue.add(Iterators.decorate(iterator1));
        queue.add(Iterators.decorate(iterator2));
        return join(new QueueIterator<FluentIterator<? extends T>>(queue));
    }

    /**
     * @param iterator1 first iterator
     * @param iterator2 second iterator to append to the first
     * @param iterator3 third iterator to append to the first and second
     * @returns iterable wich contains elements from first, second and third iterator in this order.
     */
    public static <T> FluentIterator<T> append(Iterator<? extends T> iterator1, Iterator<? extends T> iterator2, Iterator<? extends T> iterator3) {
        Queue<FluentIterator<? extends T>> queue = new ArrayDeque<FluentIterator<? extends T>>(3);
        queue.add(Iterators.decorate(iterator1));
        queue.add(Iterators.decorate(iterator2));
        queue.add(Iterators.decorate(iterator3));
        return join(new QueueIterator<FluentIterator<? extends T>>(queue));
    }

    /**
     * @param iterator wich contains iterator
     * @returns iterator wich contains elements from each iterator contained in original iterator.
     */
    public static <T> FluentIterator<T> join(Iterator<? extends Iterator<? extends T>> sources) {
        return decorate(new JoinedIterator<T>(sources));
    }

    /**
     * @param iterator to take elements from
     * @param count number of elements to take
     * @returns iterator wich contains only first count elements from iterator.
     */
    public static <T> FluentIterator<T> take(Iterator<T> source, int count) {
        return decorate(new TakenIterator<T>(decorate(source), count));
    }

    /**
     * @param iterator to take elements from
     * @param count number of elements to take
     * @returns iterator wich contains all except first count elements from iterator.
     */
    public static <T> FluentIterator<T> drop(Iterator<T> iterator, int count) {
        while (count > 0 && iterator.hasNext()) {
            iterator.next();
            count++;
        }
        return decorate(iterator);
    }

    /**
     * @param iterator iterator to base result upon
     * @param combiner operation to combine each element of iterable with seed
     * @param seed value to base result upon
     * @returns the result of applying combiner to the seed and to each element of iterator.
     */
    public static <T, U> T fold(Iterator<U> source, T seed, BiFunction<? super U, T, T> combiner) {
        T result = seed;
        while (source.hasNext())
            result = combiner.apply(source.next(), result);
        return result;
    }

    /**
     * @param iterator iterable to take element from
     * @returns new array list wich contain all elements of this iterator.
     */
    public static <T> List<T> intoNewArrayList(Iterator<? extends T> iterator) {
        List<T> list = Collections.newArrayList();
        while (iterator.hasNext())
            list.add(iterator.next());
        return list;
    }

    /**
     * @param iterator iterable to take element from
     * @returns new tree set wich contain all elements of this iterator.
     */
    public static <T> Set<T> intoNewTreeSet(Iterator<? extends T> iterator) {
        Set<T> set = Collections.newTreeSet();
        while (iterator.hasNext())
            set.add(iterator.next());
        return set;
    }

    /**
     * @param iterator iterable to take element from
     * @returns new hash set wich contain all elements of this iterator.
     */
    public static <T> Set<T> intoNewHashSet(Iterator<? extends T> iterator) {
        Set<T> set = Collections.newHashSet();
        while (iterator.hasNext())
            set.add(iterator.next());
        return set;
    }

    private Iterators() {}
}
